import functools
import os
from pathlib import Path


@functools.lru_cache(maxsize=16)
def which(binary_name: str) -> Path:
    """
    Get absolute path of an executable given name.

    Searches in $PATH.
    The result of the function is LRU cached.

    Args:
        binary_name (str): The name of the executable binary.

    Returns:
        Path: Absolute path of the executable.

    Raises:
        RuntimeError: If the executable was not found.

    """
    possible_directories = os.get_exec_path()
    for dr in possible_directories:
        exec_path = Path(dr, binary_name)
        if exec_path.exists():
            return exec_path.absolute()

    msg = f"The executable '{binary_name}' was not found in $PATH"
    raise RuntimeError(msg)



import asyncio
import os
import pkgutil
import signal
import time
from asyncio import create_subprocess_exec, create_subprocess_shell
from pathlib import Path
from typing import Any, Dict, List, Optional, Union

from knot_resolver.utils.compat.asyncio import to_thread


async def call(
    cmd: Union[str, bytes, List[str], List[bytes]], shell: bool = False, discard_output: bool = False
) -> int:
    """Async alternative to subprocess.call()."""
    kwargs: Dict[str, Any] = {"preexec_fn": signal.pthread_sigmask(signal.SIG_UNBLOCK, signal.valid_signals())}
    if discard_output:
        kwargs["stdout"] = asyncio.subprocess.DEVNULL
        kwargs["stderr"] = asyncio.subprocess.DEVNULL

    if shell:
        if isinstance(cmd, list):
            msg = "can't use list of arguments with shell=True"
            raise RuntimeError(msg)
        proc = await create_subprocess_shell(cmd, **kwargs)
    else:
        if not isinstance(cmd, list):
            msg = "Please use list of arguments, not a single string. It will prevent ambiguity when parsing"
            raise RuntimeError(msg)
        proc = await create_subprocess_exec(*cmd, **kwargs)

    return await proc.wait()


async def readfile(path: Path) -> str:
    """Asynchronously read file on a path and return its content."""

    def readfile_sync(path: Path) -> str:
        with path.open("r", encoding="utf8") as file:
            return file.read()

    return await to_thread(readfile_sync, path)


async def writefile(path: Path, content: str) -> None:
    """Asynchronously set content of a file on path to a given string content."""

    def writefile_sync(path: Path, content: str) -> int:
        with path.open("w", encoding="utf8") as file:
            return file.write(content)

    await to_thread(writefile_sync, path, content)


async def wait_for_process_termination(pid: int, sleep_sec: float = 0) -> None:
    """
    Wait for any process (does not have to be a child process) given by its PID to terminate.

    Will wait for any process (does not have to be a child process)
    given by its PID to terminate sleep_sec configures the granularity,
    with which we should return.
    """

    def wait_sync(pid: int, sleep_sec: float) -> None:
        try:
            while True:
                os.kill(pid, 0)
                if sleep_sec == 0:
                    os.sched_yield()
                else:
                    time.sleep(sleep_sec)
        except ProcessLookupError:
            pass

    await to_thread(wait_sync, pid, sleep_sec)


async def read_resource(package: str, filename: str) -> Optional[bytes]:
    return await to_thread(pkgutil.get_data, package, filename)



import asyncio
import functools
import logging
import sys
from typing import Any, Callable, Coroutine, TypeVar

logger = logging.getLogger(__name__)

T = TypeVar("T")


async def to_thread(func: Callable[..., T], *args: Any, **kwargs: Any) -> T:
    # version 3.9 and higher, call directly
    if sys.version_info >= (3, 9):
        return await asyncio.to_thread(func, *args, **kwargs)
    # earlier versions, run with default executor
    loop = asyncio.get_event_loop()
    pfunc = functools.partial(func, *args, **kwargs)
    return await loop.run_in_executor(None, pfunc)


def async_in_a_thread(func: Callable[..., T]) -> Callable[..., Coroutine[None, None, T]]:
    async def wrapper(*args: Any, **kwargs: Any) -> T:
        return await to_thread(func, *args, **kwargs)

    return wrapper


def add_async_signal_handler(signal: int, callback: Callable[[], Coroutine[Any, Any, None]]) -> None:
    loop = asyncio.get_event_loop()
    loop.add_signal_handler(signal, lambda: asyncio.create_task(callback()))


def remove_signal_handler(signal: int) -> bool:
    loop = asyncio.get_event_loop()
    return loop.remove_signal_handler(signal)


def is_event_loop_running() -> bool:
    loop = asyncio.events._get_running_loop()  # noqa: SLF001
    return loop is not None and loop.is_running()
